#
# MIT License
#
# Copyright 2021 NXP
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import glob
import logging
import tempfile
from xml.etree import cElementTree as ET

from mcutk.apps import eclipse
from mcutk.sdk_manifest import SDKManifest
from mcutk.exceptions import ProjectNotFound, InvalidProject, ProjectParserError

class Project(eclipse.Project):
    """MCUXpresso SDK and projects parser tool."""

    PROJECT_EXTENSION = '.xml'

    @classmethod
    def frompath(cls, path):
        """Return a project instance from a given file path or directory.

        If path is a directory, it will search the project file and return an instance.
        Else this will raise mcutk.apps.exceptions.ProjectNotFound.
        """
        if not os.path.exists(path):
            raise ProjectNotFound("No such file or directory: %s" % path)

        if os.path.isfile(path) and path.endswith(cls.PROJECT_EXTENSION):
            return cls(path)

        instance = None
        for filepath in glob.glob(path + "/.project"):
            with open(filepath) as fobj:
                content = fobj.read()
                if '<name>mcux</name>' in content or 'mcuxpresso' in content:
                    instance = cls(filepath)
                    break

        if instance:
            return instance

        for filepath in glob.glob(path + "/*" + cls.PROJECT_EXTENSION):
            try:
                instance = cls(filepath)
                break
            except InvalidProject:
                pass
        else:
            raise ProjectNotFound("Not found <example>.xml")

        return instance

    def __init__(self, prjpath, sdk_root=None, **kwargs):
        """MCUXPressoIDE project constructor.

        Arguments:
            prjpath {str} -- path to <name>.xml

        Keyword Arguments:
            sdk_root {str} -- path to sdk package root, default {None} that will be loaded from xml.
        """
        self.board = None
        self._is_package = False
        self._sdk_root = sdk_root
        self._name = ''
        self._targets = None
        self._sdkmanifest = Project.SDK_MANIFEST
        self._example_id = None
        self._example_xml = None
        self._nature = 'org.eclipse.cdt.core.cnature'
        self._buildproperties = None

        super(Project, self).__init__(prjpath, **kwargs)
        # eclipse project
        self._is_package = not (prjpath.endswith('.project') or prjpath.endswith('.cproject'))

        if self._is_package:
            self._load_from_sdk_package(prjpath)
            self._properties_init()

            ET.register_namespace("ksdk", "http://nxp.com/ksdk/2.0/ksdk_manifest_v3.0.xsd")
            self._example_xml = ET.parse(prjpath)

    @property
    def is_enabled(self):
        """Identify the example if is enabled(SDK package only).
        """
        if not self.is_package:
            return True

        # check manifest: this exmaple is enabled for mcux
        example_info = self.sdkmanifest.find_example(self._example_id)
        if not example_info:
            return False
        # manifest version 3.1
        if self.sdkmanifest.manifest_version == "3.1":
            return True

        return "mcux" in example_info.get("toolchain", "")

    @property
    def sdkmanifest(self):
        """Getter for SDKMainfest object"""
        return self._sdkmanifest

    @sdkmanifest.setter
    def sdkmanifest(self, value):
        """Setter for SDKMainfest object"""
        if value is None:
            return

        if not isinstance(value, SDKManifest):
            raise ValueError("Must be a SDKManifest object")

        self._sdkmanifest = value

    @property
    def build_properties(self):
        """Return a dict contains properties for build."""
        return self._buildproperties

    @property
    def is_package(self):
        """Package project or standard eclipse project"""
        return self._is_package

    def parse_cproject(self, cpath):
        """Override default .cproject parser.
        MCUXpressoIDE saved SDK info into:
            .//storageModule[@moduleId=\"com.nxp.mcuxpresso.core.datamodels\"]
        """
        conf = super(Project, self).parse_cproject(cpath)
        storage_module = self.cproject\
            .find(".//storageModule[@moduleId=\"com.nxp.mcuxpresso.core.datamodels\"]")
        if storage_module is not None:
            self.board = storage_module.find(".//boardId").text
        return conf

    def _load_from_eclipse_project(self, path):
        """Load from Eclipse C/C++ project"""
        self.parse(path)

    def _load_from_sdk_package(self, path):
        """Load from SDK <app>.xml and *_manifest*.xml.
            1. Parse <app>.xml to get manifest.xml,
            2. Get related information from manifest.
        """

        self._targets = self._conf.keys()
        try:
            xmlroot = ET.parse(path).getroot()
            example_node = xmlroot.find('./example')
            if example_node is None:
                raise InvalidProject('Unable to find <example> node. %s' % path)
        except (IOError, ValueError) as err:
            raise InvalidProject('parse xml error: %s ,path:%s' % (err, path))

        self._example_id = example_node.attrib.get('id')
        # in some situation, the name attribute is not too simple
        # that is not full project name for mcux, we have to use a workaround
        # to get project name from path.
        example_name = example_node.attrib.get('name')
        xml_name = os.path.basename(path).replace('.xml', '')
        if example_name in xml_name:
            self._name = example_name
        else:
            self._name = xml_name

        try:
            self._nature = example_node.find('projects/project[@nature]').attrib.get('nature')
        except:
            pass

        if not self._example_id:
            raise ProjectParserError('None id in exmaple node! %s'%self.prjpath)

        self._conf = {
            'Debug': self._example_id + '/Debug/',
            'Release': self._example_id + '/Release/'
        }
        if self.sdkmanifest:
            return

        # find sdk manifest from parent folders
        self.sdkmanifest = SDKManifest.find_from_parents(self.prjdir)
        if self.sdkmanifest:
            Project.SDK_MANIFEST = self.sdkmanifest
            return

        raise ProjectParserError("Unable to find SDK Manifest!")

    def _properties_init(self):
        """Init build properties variable.

        sdk.location = D:/Users/B46681/Desktop/SDK_2.0_MK64FN1M0xxx12-drop4
            This is the location where your SDK have been downloaded.
            You can use either zip or folder containing the SDK
            Please remember that if you want to create linked resources into your
            project(i.e. standalone = false) you need to use a folder instead of a zip.
            NOTE: on Windows you have to use "//" or "/".

        example.xml = D:/../hello_world/mcux/hello_world.xml
            If adding the "example.xml" property, the examples are retrieved from that
            specific file and shall valid against the used SDK
            NOTE: on Windows you have to use "//" or "/".

        nature = org.eclipse.cdt.core.cnature
            This represents the nature of your project (i.e. C or C++)
            It can be:
                - org.eclipse.cdt.core.cnature for C projects
                - org.eclipse.cdt.core.ccnature for C++ projects
                (Please remember that the example your're going to create shall support the
                C++ nature)

        standalone = true
            If true, it will copy the files from the SDK, otherwise it will link them.
            Note: linked resources will be only created if the SDK is provided as a folder

        project.build = true
            If true, the project will be compiled, otherwise the project is only created.

        clean.workspace = true
            True, if you want to clear the workspace used, false otherwise

        build.all = false
            If true, all the examples from all the SDK will be created, otherwise you need
            specify the SDK name

        skip.default = false
            If true, skip the default SDKPackages folder and all its content
            Default is false

        sdk.name = SDK_2.0_MK64FN1M0xxx12
            The SDK name (i.e. the folder/file name without extension)
            NOTE: only used when build.all = false

        board.id = frdmk64f
            The board id as for the manifest definition
            NOTE: only used when build.all = false

        Other Settings:
            verbose = true
                If true, more info will be provided using stdout

            indexer = false
                If true, enable the CDT indexer, false otherwise

            project.build.log = true
                If true, show the CDT build log, false otherwise

            simple.project.name = true

        """
        self._buildproperties = {
            'sdk.location': None,
            'example.xml': None,
            'nature': 'org.eclipse.cdt.core.cnature',
            'standalone': 'true',
            'project.build': 'true',
            'clean.workspace': 'false',
            'build.all': 'false',
            'build.config': 'debug',
            'simple.project.name': 'false',
            'use.other.files': 'true ',
            'skip.default': 'true',
            'sdk.name': None,
            'board.id': '',
            'verbose': 'false',
            'indexer': 'false',
            'use.io.console': 'false',
            'project.build.log': 'true',
            'use.semihost.hardfault.handler': 'true'
        }

    def gen_properties(self, target, loc=None):
        """Return a file path for properties file.

        Arguments:
            target -- {string} target configuration
            loc -- {string} the location to place the new geneated file,
            default is None means system tempfile.

        """
        # boardid will effect workspace path
        board_ids = self.sdkmanifest.boards
        boardid = self._example_id.replace("_" + self._name, '')
        if boardid not in board_ids:
            boardid = board_ids[0]

        logging.info("SDK Manifest Version: %s", self.sdkmanifest.manifest_version)
        # There may be multiple example.xml, to use correct xml we should
        # look up the manifest file.
        example_info = self.sdkmanifest.find_example(self._example_id)
        example_xml_filename = example_info.get('example.xml')
        if example_xml_filename:
            example_xml = os.path.join(self.prjdir, example_xml_filename)
        else:
            # Compatible with old SDK package.
            example_xml = self.prjpath
        self.setproperties("example.xml", example_xml.replace('\\', '/'))
        self.setproperties("sdk.location", self.sdkmanifest.sdk_root.replace('\\', '/'))
        self.setproperties("nature", self.nature)
        self.setproperties("sdk.name", self.sdkmanifest.sdk_name)
        self.setproperties("board.id", boardid)
        self.setproperties("build.config", target)

        with tempfile.NamedTemporaryFile(dir=loc, delete=False, prefix="mcux_", mode='w') as f:
            for per_property, value in self.build_properties.items():
                f.writelines("{0} = {1}\r\n".format(per_property, value))
            properties_file = f.name

        logging.debug('properties file: %s', properties_file)
        return properties_file

    def setproperties(self, attrib, value):
        """ Set the value of self._buildproperties"""

        self._buildproperties[attrib] = value

    def _get_defines(self, target=None):
        """ Get cc defines nodes in project, target is dummy param
        Return the cc defines ET.element
        """
        # Get the element including all macro defines
        # Format: <option id="gnu.c.compiler.option.preprocessor.def.symbols"><value>CPU_LPC845M301JBD48</value></option>
        cc_defines = self._example_xml.getroot()\
            .find(".//option[@id='gnu.c.compiler.option.preprocessor.def.symbols']")
        assert isinstance(cc_defines, ET.Element)
        return cc_defines

    def get_defines(self, target=None):
        """ get cc defines, target is dummy param, return the macro string list
        """
        macros = list()
        cc_defines = self._get_defines().findall("./value")
        for node in cc_defines:
            assert isinstance(node, ET.Element)
            macro = node.text
            if macro and macro not in macros:
                macros.append(macro)
        return macros

    def add_defines(self, new, target=None):
        """ add cc defined macro in project target configuration, target is dummy param"""
        if not new:
            raise ValueError("new macro must be set")

        changed_flag = False
        defines_elem = self._get_defines()
        macros = self.get_defines()

        if new not in macros:
            new_elem = ET.SubElement(defines_elem, 'value')
            new_elem.text = new
            changed_flag = True
            logging.debug("++macro '%s' in mcux", new)

        if changed_flag:
            self._example_xml.write(self.prjpath, xml_declaration=True,
                                    method='xml', encoding='UTF-8')

    def del_defines(self, old, target=None):
        """ delete cc defined macro in project target configuration, target is dummy param"""
        if not old:
            raise ValueError("target macro must be set to delete")

        changed_flag = False
        defines_elem = self._get_defines()
        macros_elems = defines_elem.findall("./value")
        for elem in macros_elems:
            if old == elem.text:
                defines_elem.remove(elem)
                changed_flag = True
                logging.debug("--macro '%s' in mcux", old)

        if changed_flag:
            self._example_xml.write(self.prjpath, xml_declaration=True,
                                    method='xml', encoding='UTF-8')

    @property
    def nature(self):
        """Retrun used nature"""
        return self._nature

    @property
    def targets(self):
        """Return all targets name

        Returns:
            list -- a list of targets
        """
        if self._targets:
            return list(self._targets)

        return ['Debug', 'Release']

    @property
    def name(self):
        """Return the application name

        Returns:
            string --- app name
        """
        return self._name

    def to_dict(self):
        """Dump project basic info to a dict.

        Sample:
            {
                'toolchain': "iar",
                'targets': ["debug", "release"],
                'project': "C:/path/project/",
                'name': "hello_world_project"
            }
        """
        _dict = {
            'toolchain': self.idename,
            'targets': self.targets,
            'project': self.prjpath,
            'name': self.name
        }
        if self.board:
            _dict["board"] = self.board

        return _dict
