#
# MIT License
#
# Copyright 2021 NXP
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import sys
import logging
from distutils.version import LooseVersion

from pyocd import __version__ as pyocd_version
from pyocd.target import TARGET
from pyocd.core.helpers import ConnectHelper
from pyocd.flash.file_programmer import FileProgrammer
from mcutk.debugger.general import DebuggerBase
from mcutk.util import run_command

logging.getLogger('pyocd.flash.loader').setLevel(logging.INFO)
logging.getLogger('pyocd.coresight.rom_table').setLevel(logging.WARNING)

class PYOCD(DebuggerBase):
    """
    Wrap pyOCD to a standard debugger for pymcutk.

    pyOCD is an open source Python package for programming and debugging Arm Cortex-M
    microcontrollers using multiple supported types of USB debug probes. It is fully
    cross-platform, with support for Linux, macOS, and Windows.

    For more visit https://github.com/mbedmicro/pyOCD.
    """

    def __init__(self, **kwargs):
        super(PYOCD, self).__init__("pyocd", ".", **kwargs)
        self.target_override = None
        self.version = pyocd_version
        self.loose_version = LooseVersion(self.version) if self.version else LooseVersion("0.1")

    @property
    def is_ready(self):
        return self.version not in (None, '')

    def set_board(self, board):
        # workaround to clear the prefix of usbid
        if board.usbid and ":" in board.usbid:
            board.usbid = board.usbid.split(":")[-1]
        self._board = board
        self._check_overried_target(board.devicename)

    def get_session(self, **kwargs):
        """Return a pyocd Session."""
        if kwargs.get('auto_open') is False:
            kwargs['auto_open'] = False
            kwargs['open_session'] = False

        return ConnectHelper.session_with_chosen_probe(
            blocking=False,
            unique_id=self._board.usbid,
            target_override=self.target_override,
            **kwargs
        )

    def _check_overried_target(self, devicename):
        """Overried target if device name is include in pyOCD.target.TARGET.
        This is very useful if you want to use other device target for debugging.
        """
        if devicename.endswith('pack') and os.path.exists(devicename):
            self.target_override = devicename

        elif self.is_ready and devicename in TARGET:
            self.target_override = devicename

    def _get_default_session_options(self):
        """ pyocd support to config options to control
        pyocd behaviour. Guide: https://github.com/mbedmicro/pyOCD/blob/master/docs/options.md
        """
        options = {}
        # pyocd 0.23: pre_reset to deal with low power mode
        # workaround for none kinetis series
        if self._board and self._board.name.startswith("frdm"):
            options['connect_mode'] = 'under-reset'

        return options

    def _build_cmd_args(self, board=None, options=None):
        """Build command line arguments."""
        if not board:
            board = self._board

        if not options:
            options = self._get_default_session_options()

        args = list()
        if board:
            args.append("-u %s" % board.usbid)

        if self.target_override:
            if self.target_override.endswith('.pack'):
                args.append('--pack %s' % self.target_override)
            else:
                args.append('-t %s' % self.target_override)

        if options:
            for item in options.items():
                args.append("-O%s=%s" % (item[0], str(item[1]).lower()))

        return " ".join(args)

    def list_connected_devices(self):
        """List connected CMSIS-DAP devices."""

        probes = ConnectHelper.get_all_connected_probes(blocking=False)
        devices = list()
        for probe in probes:
            device = {
                'usbid': probe.unique_id,
                'name': probe.description,
                'type': 'pyocd'
            }
            devices.append(device)
        return devices

    def test_conn(self):
        """Test debugger connection."""

        if self._board is None:
            raise ValueError("board is not set")

        msg = "NoError"
        try:
            with self.get_session():
                pass
        except Exception as err:
            msg = "ConnectError: %s" % str(err)

        return msg

    def read32(self, addr):
        with self.get_session(options={'connect_mode': 'attach'}) as session:
            return session.board.target.read_memory(addr)

    def write32(self, addr, value):
        with self.get_session(options={'connect_mode': 'attach'}) as session:
            session.board.target.write_memory(addr, value)

    def erase(self, **kwargs):
        """Mass erase flash."""
        opt_args = self._build_cmd_args(options={
            "resume_on_disconnect": False,
            "allow_no_cores": True
        })
        command = "\"{py}\" -m pyocd erase --mass -v --no-wait {opts}"\
                  .format(py=sys.executable, opts=opt_args)
        logging.info(command)
        return run_command(command, stdout=True, timeout=300)

    def reset(self):
        """Always perform a hardware reset"""

        logging.info("resetting board by pyocd")
        # do not auto open probe by session it self, because we do not need to
        # init target and flash. Just use probe to perform a hardware reset.
        # This can improve the stability even if board is in lower mode.
        session = self.get_session(auto_open=False)
        if not session:
            logging.error("cannot create session with probe id: %s", self._board.usbid)
            return False

        try:
            session.probe.open()
            session.probe.set_clock(1000000)
            session.probe.connect()
            session.probe.reset()
            session.probe.disconnect()
        finally:
            session.probe.close()
        logging.info("reset done.")
        return True

    def unlock(self):
        """Unlock board."""

        logging.info("unlock board")
        try:
            with self.get_session(options={'auto_unlock': True}):
                pass
            return True
        except:
            logging.exception("exception")
            return False

    def get_gdbserver(self, **kwargs):
        """Return gdb server command. If configured board.devicename is a target
        name in pyocd.target.TARGET. This will overried the target.
        """

        board = kwargs.get('board')
        port = kwargs.get('port')

        if board is None:
            board = self._board

        if board:
            port = board.gdbport

        command = "\"{}\" -m pyocd gdbserver --no-wait".format(sys.executable)
        if port:
            command += " --port %s" % port

        if kwargs.get('erase'):
            command += " --erase %s" % kwargs.get("erase")

        if kwargs.get("script"):
            command += " --script %s" % kwargs.get("script")

        if kwargs.get("core"):
            command += " --core %s" % kwargs.get("core")

        options = self._get_default_session_options()
        # set console to disable telent server
        options['semihost_console_type'] = 'console'
        options['keep_unwritten'] = '0'

        command += " " + self._build_cmd_args(board, options)
        return command

    def flash(self, filepath, **kwargs):
        """Flash chip with filepath. (bin, hex).

        Arguments:
            erase: erase chip yes or not.
            addr: being the integer starting address for the bin file.
        """
        addr = kwargs.get('addr')
        if addr is None:
            addr = self._board.start_address

        if addr:
            logging.info("start address: %s", addr)
            if isinstance(addr, str):
                addr = int(addr, 0)

        filepath = filepath.replace("/", "/")
        options = self._get_default_session_options()
        session = self.get_session(options=options)

        if session is None:
            raise ValueError("No device available to flash")

        file_format = 'elf'
        if filepath.endswith(".bin"):
            file_format = 'bin'
        elif filepath.endswith('.hex'):
            file_format = 'hex'

        with session:
            # call registerd callback function
            self._call_registered_callback("before_load")
            programmer = FileProgrammer(session, trust_crc=True)
            programmer.program(filepath, file_format, base_address=addr)
        return 0, ''

    @property
    def default_gdb_commands(self):
        """Return default pypcd gdb commands."""
        commands = [
            "target remote localhost:{gdbport}",
            "monitor reset --halt",
            "monitor halt",
            "load",
            "{PC_SP}",
            "q"
        ]
        return "\n".join(commands)

    def __str__(self):
        return "pyocd-%s" % self.version

    @staticmethod
    def get_latest():
        """Return pyocd.Debugger instance."""
        debugger = PYOCD()
        return debugger if debugger.is_ready else None
